#!/usr/bin/python2.4
import random
import gtk, gobject, cairo, rsvg, math

# Some flies
# Karl Lattimer <karl@qdh.org.uk>

def _rtop(x, y, deg=False):
    if deg:
        angle = 180.0 * math.atan2(y, x) / math.pi
        length = math.hypot(x, y)
    else: 
        angle = math.atan2(y, x)
        length = math.hypot(x, y)
    return (angle, length)

def _ptor(angle, length, deg=False):
    if deg: 
        angle = math.pi * angle / 180.0 
    x = length * math.cos(angle)
    y =  length * math.sin(angle)
    return x, y

class Fly:
    def __init__(self, cr, num_of_flies, width=None, height=None, max_speed=None, x=None, y=None, velocity_x=None, velocity_y=None):
        if not x:
            self.x = random.randint(0, width)
        else:
            self.x = x
        if not y:
            self.y = random.randint(0, height)
        else:
            self.y = y
        if not velocity_x:
            self.velocity_x = random.randint(-1*max_speed, max_speed)
        else:
            self.velocity_x = velocity_x
        if not velocity_y:
            self.velocity_y = random.randint(-1*max_speed, max_speed)
        else:
            self.velocity_y = velocity_y
        self.tail_x = 0
        self.tail_y = 0       
        self.cr = cr
        self.color = None
        self.num_of_flies = num_of_flies
        self.max_speed = max_speed
        self.neighbours = []
        self.eyes = None
        self.feed = 0
        self.fresheyes = False
        self.AddNeighbour()
        self.AddNeighbour()
        
    def AddNeighbour(self, fly=None):
        if len(self.neighbours) > 4:
            return
        if not fly:
            fly = random.randint(0, self.num_of_flies-1)
        self.neighbours.append(fly)
    
    def DelNeighbour(self, fly):
        try:
            #fly = self.neighbours[fly-1]
            self.neighbours.remove(fly)
        except:
            pass
    
    def SetNumOfFlies(self, num_of_flies):
        if num_of_flies:
            self.num_of_flies = num_of_flies
    
    def ReplaceNeighbour(self, old_fly, new_fly=None, num_of_flies=None ):
        #print "Proximity control"
        self.SetNumOfFlies(num_of_flies)
        self.DelNeighbour(old_fly)
        self.AddNeighbour(new_fly)
    
    def ReplaceRandom(self, num_of_flies=None, hit=None):
        if num_of_flies:
            self.num_of_flies = num_of_flies
        if not hit:
            hit = random.randint(0, len(self.neighbours))
        for index, neighbour in enumerate(self.neighbours):
            if index == hit:
                self.neighbours[index] = random.randint(0, self.num_of_flies-1)
    
    def Move( self, x=None, y=None ):
        if not x:
            x = self.x+self.velocity_x
        if not y:
            y = self.y+self.velocity_y
        self.tail_x = self.x
        self.tail_y = self.y
        self.x = x
        self.y = y

    def Eat(self):
        print "Feeding %d" % self.feed
        self.feed = self.feed + 1
    
    def Hunt(self, x, y):
        
        for eye in self.eyes:
            (direction, fov) = eye
            (angle, length) = _rtop(self.x - self.tail_x,
                                    self.y - self.tail_y, 
                                    True)
        
            self.lea = math.radians(angle + direction - fov/2)
            self.rea = math.radians(angle + direction + fov/2)

            (angle, length) = _rtop(x - self.x , y - self.y)
            if int(length) < 8 and self.lea < angle and self.rea > angle:  
                self.Eat()   
                return True
            else:
                return False
            
            dist_x = x - self.x
            dist_y = y - self.y
            (angle, length) = _rtop(self.x - x, self.y - y)
        
            if int(length) < self.view_distance and self.lea < angle and self.rea > angle:            
                direction = -6
            else:
                direction = 2
            if dist_x != 0:
                vx = self.velocity_x 
                self.velocity_x = vx + direction * (math.fabs(dist_x)/dist_x)
            vx = self.velocity_x
            if self.velocity_x > self.max_speed:
                self.velocity_x = (math.fabs(vx)/2)
            if self.velocity_x < (-1*self.max_speed):
                self.velocity_x = (-1*(math.fabs(vx)/2))
           
            if dist_y != 0:
                vy = self.velocity_y
                self.velocity_y = vy + direction * (math.fabs(dist_y)/dist_y)
            vy = self.velocity_y
            if self.velocity_y > self.max_speed:
                self.velocity_y = (math.fabs(vy)/2)
            if self.velocity_y < (-1*self.max_speed):
                self.velocity_y = (-1*(math.fabs(vy)/2))
        
        self.Move()
        
            
    def AquireTarget(self, cr, x, y, color=None, callback=None):
        if self.x == x and self.y == y:
            return
        if color:
            (sr, sg, sb) = color
        self.cr = cr
        (r, g, b) = self.color
        (angle, length) = _rtop(x - self.x , y - self.y)
        #print "------"
        #print angle
        #print self.lea
        #print self.rea
        #print length
        self.cr.save()
        #self.cr.arc(self.x, self.y, self.view_distance, 0, 2.0 * math.pi)
        #self.cr.set_line_width( 2 )
        #self.cr.set_source_rgba(.8, 0, 0, .4)    
        #self.cr.stroke()
        for eye in self.eyes:
            (direction, fov) = eye
            (angle, length) = _rtop(self.x - self.tail_x,
                                    self.y - self.tail_y, 
                                    True)
        
            self.lea = math.radians(angle + direction - fov/2)
            self.rea = math.radians(angle + direction + fov/2)
            (angle, length) = _rtop(x - self.x , y - self.y)
            
            #(angle, length) = _rtop(self.x - x, self.y - y)
   
            if int(length) < self.view_distance and self.lea < angle and self.rea > angle:
                self.cr.arc(x, y, 8, 0, 2.0 * math.pi)
                self.cr.set_line_width( 1 )
                if color:
                    self.cr.set_source_rgba(sr, sg, sb, .1)
                else: 
                    self.cr.set_source_rgba(r, g, b, .1)
                self.cr.fill_preserve()
            
                if color:
                    self.cr.set_source_rgba(sr, sg, sb, .5)
                else: 
                    self.cr.set_source_rgba(r, g, b, .5)
                self.cr.stroke()
#        (angle, length) = _rtop(self.x - self.tail_x,
#                                self.y - self.tail_y, 
#                                True)
                self.cr.move_to( self.x, self.y)        
                self.cr.line_to( x, y )
                self.cr.stroke()
                if callback:
                    return callback(x,y)
                   
            #print "Sight"
            #self.cr.explode()
        self.cr.restore()
        
    def Render(self, cr, hasprey=False):
        self.cr = cr
        self.cr.save()
        (r, g, b) = self.color
        self.cr.set_source_rgb(r, g, b)
        
        # Draw the head
        self.cr.arc(self.x, self.y, 4, 0, 2.0 * math.pi)
        self.cr.set_line_width( 3 )
        self.cr.fill()
        self.cr.restore()
        
        self.cr.save()
        # Draw the tail
        if self.tail_x == 0:
            self.Move()
        (angle, length) = _rtop(self.tail_x - self.x,
                                self.tail_y - self.y)
        self.length = (length/2)+6
        (tail_x, tail_y) = _ptor(angle, self.length)
        self.cr.set_source_rgb(r, g, b)
        self.cr.set_line_width( 3 )
        self.cr.move_to( tail_x+self.x, tail_y+self.y )        
        self.cr.line_to( self.x, self.y )
        self.cr.stroke()
        self.cr.restore()
        
        self.cr.save()
        if not self.eyes or self.fresheyes:
            self.eyes = []
         
        if not hasprey:
            self.view_distance = 40
            self.cr.set_source_rgba(r, b, g, .05)
            if len(self.eyes) < 2:
                self.eyes.append((-90, 120))
                self.eyes.append((90, 120))
            if len(self.eyes) > 2:
                self.eyes.append((-90, 120))
                self.eyes.append((90, 120))
        else:
            self.view_distance = 70
            self.cr.set_source_rgba(b, g, r, .07)
            if len(self.eyes) < 1:
                self.eyes.append((0, 40))
            if len(self.eyes) > 1:
                self.eyes = []
                self.eyes.append((0, 40))
         
        for eye in self.eyes:
            # Draw the visual range
            (direction, fov) = eye
            (angle, length) = _rtop(self.x - self.tail_x,
                                    self.y - self.tail_y, 
                                    True)
        
            self.lea = math.radians(angle + direction - fov/2)
            self.rea = math.radians(angle + direction + fov/2)
        
            (self.lex, self.ley) = _ptor(self.lea, self.view_distance)
            (self.rex, self.rey) = _ptor(self.rea, self.view_distance)
                
            self.cr.move_to( self.x, self.y)
            self.cr.line_to( self.lex+self.x, self.ley+self.y )
        
            self.cr.arc(self.x, self.y, self.view_distance, self.lea, self.rea)
        
            self.cr.move_to( self.rex+self.x, self.rey+self.y)        
            self.cr.line_to( self.x, self.y )
            self.cr.fill_preserve()
            self.cr.stroke()
        self.cr.restore()
        
        
class Swarm:
    def __init__(self, flies, behavior, rect, color, cr):
        (impact_velocity, max_speed, grow_rate, decay_rate, fly_type, size, repel) = behavior
        (self.c_red, self.c_green, self.c_blue) = color
        (self.x, self.y, self.width, self.height) = rect
        self.fly_type = fly_type
        self.cr = cr
        self.grow_rate = grow_rate
        self.decay_rate = decay_rate
        self.num_of_flies = 0
        self.flies = []
        self.impact_velocity = impact_velocity
        self.max_speed = max_speed
        self.stale_ticks = 1000
        self.x_grav = 2
        self.y_grav = 2
        self.fresh_neighbour = self.stale_ticks
        self.repel_distance = repel
        self.flush = False
        self.feed_volume = 0
        self.size = size
        self.predator = None
        self.prey = None
        self.prey_volume = None
        self.type = type
        for fly in range(flies):
            self.AddFly()

    def AddFly( self ):
        self.num_of_flies = self.num_of_flies + 1
        self.flies.append( Fly(self.cr, self.num_of_flies, self.width, self.height, self.max_speed) )

    def DelFly( self, fly=None ):
        if not fly:
            fly = self.flies[random.randint(0, self.num_of_flies-1)]
        self.num_of_flies = self.num_of_flies - 1
        self.flies.remove(fly)

    def SetChain(self, predator, prey, prey_volume):
        self.predator = predator
        self.prey = prey
        self.prey_volume = prey_volume
        
    def FlushRefresh(self):
        global cr
        self.cr = cr
        self.cr.save()
        self.cr.set_source_rgb(.3, 0, .1)
        self.cr.rectangle(0, 0, self.width, self.height)
        self.cr.fill()
        self.cr.restore()
        self.cr.save()
        self.flush = True
        cr = self.cr
        self.Refresh()
        
    def Refresh(self):
        global cr
        self.cr = cr
        gen = False
        if len(self.flies) < 1:
            return
        #idx = 1 + random.randint(0, (self.num_of_flies-1)/2)
        for self.fly,fly in enumerate(self.flies):
            hint = float(self.fly)/float(self.num_of_flies*2)
            self.fly = self.fly - 1
            self._proximity()
            #self._evade()
            self._grav()
            self._hunt()
            
            #for target,fly in enumerate(self.flies[self.fly].neighbours):
            #    self.flies[self.fly].AquireTarget(self.cr, self.flies[target-1].x, self.flies[target-1].y)
                
            # Births and deaths, should be moved elsewhere
            
                
            if self.fresh_neighbour == 0:
                self.fresh_neighbour = self.stale_ticks
                self.flies[self.fly].ReplaceRandom(self.num_of_flies)
                print self.fly_type + " refreshing neighbours"
            else:
                self.fresh_neighbour = self.fresh_neighbour - 1 
                
            if self.prey:
                hasprey = True
            else:
                hasprey = False 
            
            if not hasprey:
                if self.fresh_neighbour % self.grow_rate == 0:
                    print self.fly_type + " breeding"
                    self.AddFly()
                    
            if (self.fresh_neighbour) % self.decay_rate == 0:
                self.DelFly(self.flies[0])
                print self.fly_type + " dying"
                continue                    
            
            if not self.flies[self.fly].color:
                self.flies[self.fly].color = (self.c_red + hint, self.c_green + hint, self.c_blue + hint)
            
            self.flies[self.fly].Render(self.cr, hasprey)
            
        # Overpopulation
        if self.num_of_flies > 30:
            self.DelFly()
            print self.fly_type + " overpopulated"
            
    def _grav(self):
	"""
	Fly gravity, gravity point is currently fixed at a distance of 
	two thirds width from centre. This is boxy and highly inaccurate 
	and just plain pants.

	A better way would be that gravity is determined as a function
	of the polar distance from the gravity target.

	The gravity target could then be determined by some other means

	"""
        if self.flies[self.fly].x < self.width/3:
            self.flies[self.fly].velocity_x = self.flies[self.fly].velocity_x + random.randint(0, self.x_grav)
        if self.flies[self.fly].x > 2*(self.width/5):
            self.flies[self.fly].velocity_x = self.flies[self.fly].velocity_x - random.randint(0, self.x_grav)
        if self.flies[self.fly].y < self.height/3:
            self.flies[self.fly].velocity_y = self.flies[self.fly].velocity_y + random.randint(0, self.y_grav)
        if self.flies[self.fly].y > 2*(self.height/5):
            self.flies[self.fly].velocity_y = self.flies[self.fly].velocity_y - random.randint(0, self.y_grav)
                
    def _gen_birth(self):
        pass
    def _gen_death(self):
        pass
    def _evade(self):
        if self.flies[self.fly].x > (self.width - 20):
            self.flies[self.fly].velocity_x =  -1 * random.randint(0, self.impact_velocity)
        elif self.flies[self.fly].x < 20:
            self.flies[self.fly].velocity_x = random.randint(0, self.impact_velocity)

        if self.flies[self.fly].y > (self.height - 20):
            self.flies[self.fly].velocity_y = -1 *random.randint(0, self.impact_velocity)
        elif self.flies[self.fly].y < 20:
            print self.fly_type + " FEEDING"
            self.prey.DelFly(fly) 
            self.flies[self.fly].velocity_y =  random.randint(0, self.impact_velocity)
        
    def _hunt(self):
        if not self.prey:
            return
        ate = False
        for target,fly in enumerate(self.prey.flies):
            try:
                ate = self.flies[self.fly].AquireTarget(self.cr, self.prey.flies[target-1].x, 
                                                  self.prey.flies[target-1].y, 
                                                  (0.9,0,0), 
                                                  self.flies[self.fly].Hunt)
            except:
                pass
            if ate:
                self.feed_volume = self.feed_volume + 1
                print self.fly_type + " Feeding"
                self.prey.DelFly(fly)
                print self.prey.fly_type + " Eaten"
                if self.feed_volume > self.prey_volume:
                    self.AddFly()
                    print self.fly_type + " Spawning"
                    self.feed_volume = 0
                    
    
    def _proximity(self):
	"""
        If the neighbours of this flys neighbours are closer then repatriate them
        to this fly.

	The implementation of this can be improved greatly
	"""
        for index, neighbour in enumerate(self.flies[self.fly].neighbours):
            if neighbour > self.num_of_flies - 1:
                self.flies[self.fly].ReplaceRandom(self.num_of_flies, index)
                continue
            distance = self._distance(self.flies[neighbour], self.flies[self.fly])
            for index, next_neighbour in enumerate(self.flies[neighbour].neighbours):
                if next_neighbour > self.num_of_flies - 1:
                    self.flies[self.fly].ReplaceRandom(self.num_of_flies, index)
                    continue
                next_distance  = self._distance(self.flies[next_neighbour], self.flies[self.fly]) 
                if next_distance > distance and next_neighbour != self.fly:
                    change = True
                    for t_neighbour in self.flies[self.fly].neighbours:
                        if t_neighbour == next_neighbour:
                            change = False
                    if change:
                        self.flies[self.fly].ReplaceNeighbour(neighbour, next_neighbour)
                        
            dist_x = self.flies[neighbour].x - self.flies[self.fly].x
            dist_y = self.flies[neighbour].y - self.flies[self.fly].y
            
            if math.fabs(dist_x) < self.repel_distance or math.fabs(dist_y) < self.repel_distance:
                direction = -1
            else:
                direction = 1
            if dist_x != 0:
                vx = self.flies[self.fly].velocity_x
                self.flies[self.fly].velocity_x = vx + direction * (math.fabs(dist_x)/dist_x)
            vx = self.flies[self.fly].velocity_x
            if self.flies[self.fly].velocity_x > self.max_speed:
                self.flies[self.fly].velocity_x = (math.fabs(vx)/2)
            if self.flies[self.fly].velocity_x < (-1*self.max_speed):
                self.flies[self.fly].velocity_x = (-1*(math.fabs(vx)/2))
            if dist_y != 0:
                vy = self.flies[self.fly].velocity_y
                self.flies[self.fly].velocity_y = vy + direction * (math.fabs(dist_y)/dist_y)
            vy = self.flies[self.fly].velocity_y
            if self.flies[self.fly].velocity_y > self.max_speed:
                self.flies[self.fly].velocity_y = (math.fabs(vy)/2)
            if self.flies[self.fly].velocity_y < (-1*self.max_speed):
                self.flies[self.fly].velocity_y = (-1*(math.fabs(vy)/2))
            self.flies[self.fly].Move()
            
    def _collission(self):
        pass
    def _distance( self, a, b ):
        # I'm sure pythagoras is released under GPL
        x = a.x - b.x
        y = a.y - b.y
        return int(math.hypot(x, y))    
        
class FliesWidget(gtk.DrawingArea):
    __gsignals__ = { "expose-event": "override" }
    
    def do_expose_event( self, event ):
        global swarma
        global swarmb
        global cr
    
        self.width = event.area.width
        self.height = event.area.height
        self.cr = self.window.cairo_create()
        self.MyExpose()
        
    def MyExpose(self):
        global swarma
        global swarmb
        global swarmc
        global pop_string
        global cr
        
        behavior = (3, 9, 250, 250, "red", 5, 10)
        color = (.7, 0, .3)
        rect = (0, 0, self.width, self.height)
        if not swarma:
             swarma = Swarm(10, behavior, rect, color, self.cr )
        else:
            swarma.width =  self.width
            swarma.height = self.height
            swarma.cr = self.cr
        
        swarma.FlushRefresh()
        
        behavior = (1, 5, 150, 2000, "blue", 3, 4)
        color = (.1, .3, .6)
        rect = (0, 0, self.width, self.height)
        if not swarmb:
             swarmb = Swarm(5, behavior, rect, color, self.cr )
        else:
            swarmb.width = self.width
            swarmb.height = self.height
            swarmb.cr = self.cr
            
        swarmb.Refresh()
        """
        behavior = (6, 12, 6, 12, "green", 2)
        color = (.1, .7, .2)
        rect = (0, 0, self.width, self.height)
        if not swarmc:
             swarmc = Swarm(10, behavior, rect, color, self.cr )
        else:
            swarmc.width = self.width
            swarmc.height = self.height
            swarmc.cr = self.cr
            """
        swarma.SetChain( None, swarmb, 2 )
        swarmb.SetChain( swarma, None, 0 )
        #swarmc.SetChain( swarmb, None, None )
        #swarmc.Refresh()
        reds = "Reds %d " % swarma.num_of_flies
        blues = "Blues %d " % swarmb.num_of_flies
        #greens = "Greens %d" % swarmc.num_of_flies
        old_popstring = pop_string
        pop_string = reds + blues #+ greens
        if pop_string != old_popstring:
            print pop_string
        gobject.timeout_add(1000/15, self.MyExpose)
        
if __name__ == "__main__":
    print "Starting"
    swarma = None
    swarmb = None
    swarmc = None
    pop_string = "Init"
    window = gtk.Window()
    window.connect("delete-event", gtk.main_quit)
    dc = FliesWidget()
    
    window.resize(1000, 700)
    dc.show()
    window.add(dc)
    
    window.present()
    cr = dc.window.cairo_create()
    gtk.main()
