import sys, os
import gtk, gobject

from const import *

from preferences import preferences
from utils import Download
from utils import printNestedStuff
from log import log

from xml.sax import make_parser
from xml.sax.handler import ContentHandler


class PackListParse( ContentHandler ):
	def __init__( self, packs ):
		self.in_app = False
		self.in_repo = False
		self.in_pack = False
		self.in_wine = False
		self.v_data = []
		self.installersize = None
		self.installedsize = None
		self.w_version = None
		self.w_variant = None
		self.w_rating = None
		ContentHandler.__init__( self )
		self.packs = packs # [{},{}]
		self.data = []
		self.version = None
		self.pack = {} # name, text
		self.versions = [] # version, text
		self.repo = [] # [reponame, repouri], should be a dic
		while len( self.repo ) < 2:
			self.repo.append( "" )
			
	def startElement( self, name, attrs ):
		if name == "Application":
			self.in_app = True
		elif self.in_app and name == "pack":
			self.in_pack = True
			self.version = attrs.get( 'version' )
		elif self.in_pack and name == "wine":
			self.in_wine = True
			self.w_version = attrs.get( 'version' )
			self.w_variant = attrs.get( 'variant' )
			self.w_rating = attrs.get( 'rating' )
		elif name == "repo":
			self.in_repo = True
			
	def endElement( self, name ):
		if name == "Application":
			self.in_app = False
			if "ms-license" not in self.pack:
				self.pack["ms-license"] = False
			self.pack["versions"] = self.versions
			self.pack["repo"] = self.repo
			self.packs.append( self.pack )
			self.data = []
			self.version = None
			self.versions = []
			self.pack = {}
			return
		elif name == "repo":
			self.in_repo = False
			
		text = ''.join( self.data )
		
		if self.in_app and name == "pack":
			# Backward compatible for the time being
			if self.v_data == None:
				self.v_data = text
			# v_data is constructed from elements inside pack elements
			self.versions.append(self.v_data)
			self.v_data = []
			self.in_pack = False
			
		elif self.in_wine and name == "installersize":
			self.installersize = text
		elif self.in_wine and name == "installedsize":
			self.installedsize = text
		elif self.in_pack and name == "wine":
			self.v_data.append ( ( self.version,
			                       self.w_version, 
			                       self.w_variant, 
			                       self.w_rating, 
			                       self.installersize,
			                       self.installedsize,
			                       text ) )
			self.w_version = None
			self.w_variant = None
			self.w_rating = None
			self.installersize = None
			self.installedsize = None
			self.in_wine = False
		elif self.in_repo and name == "name":
			self.repo[0] = text
		elif self.in_repo and name == "uri":
			self.repo[1] = text
		elif self.in_app:
			self.pack[name] = text
		self.data = []
		
	def characters( self, string ):
		self.data.append( string.rstrip() )
		
	def error( self, exception ):
		log.Error( "%s\n" % exception )

class PackList: 
	def __init__( self, repo=None, fresh=False ):
		self.available = []
		self.packs = []
		self.installed = []
		
		packlist_path = preferences["wineroot"] + "/wine-doors/packlists/"
		if not os.path.isdir( packlist_path ):
			log.Debug(packlist_path)
			os.makedirs(packlist_path);
			self.tmp = False
			self.fresh = False
		else:
			self.tmp = False
			self.fresh = fresh
		if not repo:
			pass  
		elif repo.startswith ( "http://" ):
			self.Update( repo )
		elif os.path.isfile( repo ):
			self.packs = self.Parse( repo )
			self.packrepo = None
			self.packfile = repo
		elif repo:
			self.Update( repo )
			
		installed_list = packlist_path + "installed.packlist.xml"
		
		if os.path.isfile( installed_list ):
			self.installed = self.Parse( installed_list )
	
	def GetRepoURI ( self, repo_name ):
		"""
		Lookup the repository URI from the name
		"""
		repo_uri = ""
		for item in preferences.repositories:
			if ''.join( item[0:1] ) == repo_name:
				repo_uri = ''.join( item[1:] )
		return repo_uri
		
	def GetRepoName ( self, repo_uri ):
		"""
		Look up the repository name from the URI
		"""
		repo_name = None
		for item in preferences.repositories:
			if ''.join( item[1:] ) == repo_uri:
				repo_name = ''.join( item[0:1] )
		return repo_name
		
	def Parse( self, xmlfile ):
		"""
		Invoke a SAX parser for PackList XML
		"""
		if not os.path.isfile ( xmlfile ):
			log.Error( "No such file : " + xmlfile )
		handler = PackListParse( self.packs )
		parser = make_parser()
		parser.setContentHandler( handler )
		parser.parse( xmlfile )
	
	def Update( self, repo ):
		"""
		Retrieve an XML list of application packs and their URLs
		"""
		repo_uri = self.GetRepoURI ( repo )
		supplied_repo = repo
		if not repo_uri:
			log.Debug( repo + " is not a valid repository" )
			log.Debug( "  attempting a reverse lookup" )
			repo = self.GetRepoName( repo )
			repo_uri = self.GetRepoURI( repo )
			
		if not repo:
			log.Error( "Repo update failed for " + supplied_repo )
			
		if not repo.startswith ( "http://" ):
			repo_uri = repo_uri + "/packlist.xml.gz"
		
		log.Debug( "Checking for available applications in " + repo )
		
		if self.tmp:
			local_file = "/tmp/" + repo +".xml"
		else:
			local_file = preferences["wineroot"] + "/wine-doors/packlists/"
			local_file += repo + ".xml"
		
		Download( repo_uri, local_file + ".gz", not self.fresh )
		self.packrepo = repo
		log.Debug( local_file )
		self.Parse( local_file )
		
		try:
			len( self.available )
			self.available += self.packs
		except:
			self.available = self.packs
		
		if self.tmp:
			os.remove( local_file )
			
	def UpdateAll ( self ):
		for item in preferences.repositories:
			self.Update( ''.join( item[0:1] ) )
			self.available += self.packs

	def Save( self, tofile = "installed.packlist.xml" ):
		""" 
		Save to installed.packlist.xml 
		"""
		f = open( tofile, "w" )
		backup = sys.stdout
		sys.stdout = f
		print "<PackList version=\"0.1\">"
		for pack in self.installed:
			print "<ApplicationPack>"
			for k, v in pack.iteritems():
				if k == "version":
					for ver, url in v.iteritems():
						print "<packuri version=\"%s\">%s</packuri>" % ( ver, url )
				else:
					print "<%s>%s</%s>" % ( k, v, k )
			print "</ApplicationPack>"
		print "</PackList>"
		f.close()
		sys.stdout = backup

	def __contains__( self, key ):
		#FIXME: This is not the right way to do it... why not?
		index = self.installed.index( key )
		if index:
			return True
		return False
		
def usage():
	print "Unit testing of pack.py"
	print "Usage: python pack.py <class> <file>"
	print "  <class> one of PackListParse, PackList"
	print "  <file> file to feed into the class as a unit"	
	
if __name__ == "__main__":
	if len( sys.argv ) == 3:
		if sys.argv[1] == "PackList":
			print "Testing Class: PackList"
			print "  Parsing packlist from " + sys.argv[2]
			packlist.Update ( sys.argv[2] )
			print 
			printNestedStuff( packlist.packs )
		elif sys.argv[1] == "PackListParse":
			print "Testing Class: PackListParse"
			print "  Parsing packlist xml from " + sys.argv[2]
			xpacks = []
			xhandler = PackListParse( xpacks )
			xparser = make_parser()
			xparser.setContentHandler( xhandler )
			xparser.parse( sys.argv[2] )
			printNestedStuff( xpacks )
		else:
			print sys.argv[1] + " is not a valid class"
			usage()	   
	else:
		usage()
else:
	packlist = PackList()  
