import gtk, gobject
import os, re, time
import ui

from const import *

from application import ApplicationPack
from preferences import preferences
from packlist import packlist
from wine import wine
from log import log

PACK_NAME, PACK_SHORTNAME = range ( 2 )

class Queue:
	def __init__( self, action = False ):
		self.queue = []
		if action:
			self.AddBaseLibraries()

	def Add( self, action, pack=None, version=None, start=False ):
		if start:
			self.queue[0:0] = [( action, pack, version )]
		else:
			self.queue.append ( ( action, pack, version ) )

		if action == INSTALL:
			app_pack = ApplicationPack( DEPS, pack )
			if app_pack.status:
				for item in app_pack.requires:
					self.Add( INSTALL, item, None, True )

	def SetProgress( self, index = 0 ):
		total = len( self.queue )
		if index >= total: 
			return
		
		action, pack = self.queue[index]
		if action == DRIVE:
			name = "Creating wine drive"
		else:
			data = pack['name']
			if "default_version" in pack:
				version = pack['default_version']
			else:
				version = ALL
			name = "Installing Application Pack %s %s" % ( data, version )
		
		percentage = float( index ) / float ( total )
	   
		# FIXME: UI Code should be moved to ui.py #378 
		ui.window['progressbar'].set_text( name )
		ui.window['progressbar'].set_fraction( percentage )
		
	def Execute( self , FirstRun=False ):
		for item in self.queue:
			action, pack, version = item
				
			if action == DRIVE:
				log_text = "Creating wine drive"
			else:
				log_text  = "Installing Application: "
				log_text += "%s %s" % ( pack['name'], version_txt )
				if not version:
					log_text += " all versions"
				else:
					log_text += " version " + version
					
			log.Debug( log_text )
			if action == INSTALL:
				app_pack = ApplicationPack( "install", pack, version )
				if app_pack.status:
					log.Error( "Installation failed: " + log_text )
			elif action == DRIVE:
				wine.CreateDrive()
			self.SetProgress( self.queue.index( item ) + 1 )
			yield True

		if FirstRun:
			ui.window['wn_firstrun'].hide()
			gtk.main_quit()
		wine.Reboot()
		del( self.queue )
		yield False
	
	# FIXME: UI Code should be moved to ui.py #378 
	# Moved to ui.PackTree.CreateModel
	# DEPRECATED
	def CreateTreeModel ( self ):
		model = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_STRING )
		
		for item in self.queue:
			action, pack = item
			iter = model.append()
			model.set( iter, PACK_NAME, pack["name"], 
							 PACK_SHORTNAME, pack['shortname'] )
		return model
	
	def AddBaseLibraries( self ):
		self.Add( DRIVE )
		packlist.Update( "System Base" )
		for pack in packlist.packs:
			licenserqd = pack["ms-license"]
			if not licenserqd or ( preferences['licensed'] and licenserqd ):
				self.Add( INSTALL, pack["shortname"] )
			
queue = Queue()